@@ -51,7 +51,6 @@ class HttpConn(object):
         self.args  = hsrv.args  # mypy404
         self.E  = self.args.E
         self.asrv  = hsrv.asrv  # mypy404
-        self.cert_path = hsrv.cert_path
         self.u2fh  = hsrv.u2fh  # mypy404
         self.iphash  = hsrv.broker.iphash
         self.bans   = hsrv.bans
@@ -62,6 +61,7 @@ class HttpConn(object):
         self.ico  = Ico(self.args)  # mypy404
 
         self.t0  = time.time()  # mypy404
+        self.freshen_pwd  = 0.0
         self.stopping = False
         self.nreq  = -1  # mypy404
         self.nbyte  = 0  # mypy404
@@ -100,16 +100,17 @@ class HttpConn(object):
         self.log_func(self.log_src, msg, c)
 
     def get_u2idx(self)  :
-        # one u2idx per tcp connection;
+        # grab from a pool of u2idx instances;
         # sqlite3 fully parallelizes under python threads
+        # but avoid running out of FDs by creating too many
         if not self.u2idx:
-            self.u2idx = U2idx(self)
+            self.u2idx = self.hsrv.get_u2idx(str(self.addr))
 
         return self.u2idx
 
     def _detect_https(self)  :
         method = None
-        if self.cert_path:
+        if True:
             try:
                 method = self.s.recv(4, socket.MSG_PEEK)
             except socket.timeout:
@@ -143,7 +144,7 @@ class HttpConn(object):
         self.sr = None
         if self.args.https_only:
             is_https = True
-        elif self.args.http_only or not HAVE_SSL:
+        elif self.args.http_only:
             is_https = False
         else:
             # raise Exception("asdf")
@@ -157,7 +158,7 @@ class HttpConn(object):
             self.log_src = self.log_src.replace("[36m", "[35m")
             try:
                 ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
-                ctx.load_cert_chain(self.cert_path)
+                ctx.load_cert_chain(self.args.cert)
                 if self.args.ssl_ver:
                     ctx.options &= ~self.args.ssl_flags_en
                     ctx.options |= self.args.ssl_flags_de
@@ -211,3 +212,7 @@ class HttpConn(object):
             self.cli = HttpCli(self)
             if not self.cli.run():
                 return
+
+            if self.u2idx:
+                self.hsrv.put_u2idx(str(self.addr), self.u2idx)
+                self.u2idx = None
